----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    22:22:45 02/10/2009 
-- Design Name: 
-- Module Name:    disp_driver - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
-- this module runs at the pixel clock rate. it uses the second port(s) of the selectram+
-- instantiated by the scanline caches.
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity disp_driver is
	Port (
	tvc_clock		: in	std_logic;
	tvc_reset		: in	std_logic; --this reset is probably coming from another clock domain, needs a sync.

	--we nicly sync these outputs to other tvc components to the tvc clock.
	odd_fill_req		: out std_logic;
	even_fill_req		: out std_logic;
	reset_cntrs_req	: out std_logic; --reset odd/even address counters to display top.

	--odds are, this is a _much_faster clock than the tvc clock. (but could be slower)
	pixel_clock 	: in  STD_LOGIC;
	
	--these signals are operating in the pixel clock domain.
	next_pix_addr		: out	std_logic_vector(11 downto 0);
	--these vectors should be much wider.
	odd_next_pix_val	: in	std_logic_vector(31 downto 0); --data from the scanline cache
	even_next_pix_val	: in 	std_logic_vector(31 downto 0);
	--signals to "DAC" & monitor
	disp_pix_value 	: out 	std_logic_vector(7 downto 0);
	disp_horiz_sync 	: out 	std_logic;
	disp_vert_sync 	: out 	std_logic
	);
end disp_driver;

architecture Behavioral of disp_driver is

	component reset_sync
	Port (
	sig_in	: in	STD_LOGIC;
	sig_out	: out	STD_LOGIC;
	clock		: in	STD_LOGIC
	);
	end component;

	component std_sync is
   Port (
	sig_in	: in	STD_LOGIC;
	sig_out	: out	STD_LOGIC;
	clock		: in	STD_LOGIC;
	reset		: in	STD_LOGIC
	);
	end component;

	--signals operate at tvc clock domain
	signal odd_fill_seen			: std_logic;
	signal even_fill_seen		: std_logic;
	signal reset_cntrs_seen		: std_logic;
	signal odd_fill_int_req 	: std_logic;
	signal even_fill_int_req 	: std_logic;
	signal reset_cntrs_int_req	: std_logic;

	--signals from clock domain into tvc_clock domain
	signal odd_fill_int_req_PIXCLK 		: std_logic; --internal requests in the pixel clock domain
	signal even_fill_int_req_PIXCLK		: std_logic;
	signal reset_cntrs_int_req_PIXCLK	: std_logic;
	
	signal odd_fill_int_req_TVCCLK 		: std_logic; --internal request in the tvc clock domain
	signal even_fill_int_req_TVCCLK		: std_logic;
	signal reset_cntrs_int_req_TVCCLK	: std_logic;
	
	--signals in the pixel clock domain.
	signal hsync_counter 	: STD_LOGIC_VECTOR(15 downto 0);
	signal vsync_counter 	: STD_LOGIC_VECTOR(23 downto 0);	 
	signal v_display_pix		: std_logic; --in a vertical region of the display where we should be clocking out pixels
	signal h_display_pix		: std_logic; --in a horizontal region of the display where we should be clocking out pixels
	signal curr_drawn_scanline_is_even	: std_logic;  --only valid when v_display_pix is true
	--this must be one clock ahead of the current pixel.
	signal eo_sc_next_pix_addr : std_logic_vector(11 downto 0);

	signal synced_reset : std_logic;

begin

	odd_fill_req <= odd_fill_int_req;
	even_fill_req <= even_fill_int_req;
	reset_cntrs_req <= reset_cntrs_int_req;

	next_pix_addr <= eo_sc_next_pix_addr;
	
	rsync : reset_sync
	port map( sig_in=>tvc_reset, sig_out=>synced_reset, clock=>pixel_clock);

	odd_fill_sync : std_sync
	port map(sig_in=>odd_fill_int_req_PIXCLK, sig_out=>odd_fill_int_req_TVCCLK,
				clock=>tvc_clock, reset=>tvc_reset);

	even_fill_sync : std_sync
	port map(sig_in=>even_fill_int_req_PIXCLK, sig_out=>even_fill_int_req_TVCCLK,
				clock=>tvc_clock, reset=>tvc_reset);

	reset_cntrs_sync : std_sync
	port map(sig_in=>reset_cntrs_int_req_PIXCLK, sig_out=>reset_cntrs_int_req_TVCCLK,
				clock=>tvc_clock, reset=>tvc_reset);

	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		odd_fill_int_req <= '0';
		even_fill_int_req <= '0';
		reset_cntrs_int_req <= '0';
		odd_fill_seen <= '0';
		even_fill_seen <= '0';
		reset_cntrs_seen <= '0';
	else
		if odd_fill_int_req = '1' then --have it high only one cycle in the tvc clock domain.
			odd_fill_int_req <= '0';
		else
			if odd_fill_int_req_TVCCLK = '1' then
				if odd_fill_seen = '0' then
					odd_fill_int_req <= '1';
					odd_fill_seen <= '1';
				end if;
			else
				odd_fill_seen <= '0';
			end if;
		end if;
		
		if even_fill_int_req = '1' then --have it high only one cycle in the tvc clock domain.
			even_fill_int_req <= '0';
		else
			if even_fill_int_req_TVCCLK = '1' then
				if even_fill_seen = '0' then
					even_fill_int_req <= '1';
					even_fill_seen <= '1';
				end if;
			else
				even_fill_seen <= '0';
			end if;
		end if;
		
		if reset_cntrs_int_req = '1' then --have it high only one cycle in the tvc clock domain.
			reset_cntrs_int_req <= '0';
		else
			if reset_cntrs_int_req_TVCCLK = '1' then
				if reset_cntrs_seen = '0' then
					reset_cntrs_int_req <= '1';
					reset_cntrs_seen <= '1';
				end if;
			else
				reset_cntrs_seen <= '0';
			end if;
		end if;
	end if; --reset
	end if; --tvc_clock
	end process;

	process(pixel_clock)
		variable long_pixel_value : std_logic_vector(31 downto 0);
	begin
	if pixel_clock = '1' and pixel_clock'event then
	if synced_reset = '1' then
		--vsync_counter <= X"000000"; --HACK! commenting this out is pushing logic savings a bit far...
		--hsync_counter <= X"0000"; --HACK! commenting this out is pushing logic savings a bit far...
		v_display_pix <= '0';
		h_display_pix <= '0';
		reset_cntrs_int_req_PIXCLK <= '0';
		odd_fill_int_req_PIXCLK <= '0';
		even_fill_int_req_PIXCLK <= '0';
		disp_horiz_sync <= '1';
		disp_vert_sync <= '1';
		--disp_pix_value <= X"00"; --HACK! commenting this out is pushing logic savings a bit far...
		curr_drawn_scanline_is_even <= '1';
	else --not reset

		if vsync_counter =  X"065c20" then  --the end of the full display cycle
			vsync_counter <= X"000000";
		else
	   	vsync_counter <= vsync_counter + 1;
		end if;

		if hsync_counter = X"320" then
			hsync_counter <= X"0000";
		else
		  	hsync_counter <= hsync_counter + 1;
      end if;

		case vsync_counter is
		when X"000000"	=> --start the vsync pulse
			disp_vert_sync <= '0'; 
			reset_cntrs_int_req_PIXCLK <= '1';
			curr_drawn_scanline_is_even <= '1'; --safe to do now. we only change it below when v_display_pix == 1
		when X"000010" =>
			reset_cntrs_int_req_PIXCLK <= '0';
			odd_fill_int_req_PIXCLK <= '1';
		when X"000020" =>
			odd_fill_int_req_PIXCLK <= '0';
			even_fill_int_req_PIXCLK <= '1';
		when X"000030" =>
			even_fill_int_req_PIXCLK <= '0';		
		when X"000640" => --end the vsync pulse 
			disp_vert_sync <= '1'; 
      when X"002580" => --we are starting drawing the part of the screen where pixels are displayed 
			v_display_pix <= '1';
		when X"063ce0" => --stop drawing pixels
			v_display_pix <= '0';
		when others => null;
      end case;
	
		case hsync_counter is
		when X"0060" =>
			disp_horiz_sync <= '1';
			--decide not on our pixel source for this scanline.  HACK. we should set an invalid flag in the scanline cache for when it does complete
			--we do not display its data on the incorrect display line.  OR.. with properly designed hardware it would never happen, so we should
			--ignore the possibility (just like this test below)
			if v_display_pix = '1' then
				eo_sc_next_pix_addr <= X"000";  --ok to set here. this is protected from change when scanning a scanline by h_display_pix.
			end if;	 
		when X"0070" =>
			h_display_pix <= '1';
		when X"02F0" =>
			h_display_pix <= '0';
			if v_display_pix = '1' then
				if curr_drawn_scanline_is_even = '0' then --we just drained the odd scanline
					odd_fill_int_req_PIXCLK <= '1';
					curr_drawn_scanline_is_even <= '1'; --get ready to drain next scanline
				else
					even_fill_int_req_PIXCLK <= '1';
					curr_drawn_scanline_is_even <= '0';
				end if;
			end if;
		when X"0300" =>
			odd_fill_int_req_PIXCLK <= '0';
			even_fill_int_req_PIXCLK <= '0';
		when X"0320" =>
			disp_horiz_sync <= '0';
		when others => null;
		end case;

		-- ************ output pixel values ****************
		long_pixel_value := X"0000_0000";
		if v_display_pix = '1' and h_display_pix = '1' then
			if curr_drawn_scanline_is_even = '0' then
		   	long_pixel_value := odd_next_pix_val;
			else
				long_pixel_value := even_next_pix_val;
			end if;			
			--old pattern generator code
--			case eo_sc_next_pix_addr(5 downto 4) is
--			when "00" => 	long_pixel_value := X"c0c0c0c0";
--			when "01" =>	long_pixel_value := X"38383838";
--			when "10" =>  	long_pixel_value := X"07070707";
--			when others =>	long_pixel_value := X"00000000";
--			end case;

			eo_sc_next_pix_addr <= eo_sc_next_pix_addr + 1;
		end if;

		case eo_sc_next_pix_addr(1 downto 0) is
		when "00" =>
			disp_pix_value <=long_pixel_value(31 downto 24);
		when "01" =>
			disp_pix_value <= long_pixel_value(23 downto 16);
		when "10" =>
			disp_pix_value <= long_pixel_value(15 downto 8);
		when "11" =>
			disp_pix_value <= long_pixel_value(7 downto 0);
		when others =>
			null; --should not happen
		end case;

	end if; --reset
   end if; --pixel clock
	end process;

end Behavioral;

