library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity eppcu is
    Port (
	--the eppcu's datalines that run to the parallel port
	epp_data					: inout	std_logic_vector(7 downto 0);
	epp_write 				: in 	std_logic;
	epp_wait 				: out std_logic;
	epp_data_strobe 		: in 	std_logic;
	epp_address_strobe 	: in 	std_logic;
	epp_irq					: out	std_logic; --not used but set to zero
	--           epp_reset 		: in 	std_logic;

	--the eppcu's locic interface to the rest of the tvc		 
	tvc_clock				: in	std_logic;
	tvc_reset				: in 	std_logic;

	--interface to the command bus
	cmd_bus_data      : out std_logic_vector(31 downto 0);
   cmd_bus_addr      : out std_logic_vector(7 downto 0);
   cmd_bus_strobe    : out std_logic;
   cmd_bus_data_out  : in std_logic_vector(31 downto 0); --currently just the block reader's read value

	br_idle           : in std_logic; --interface to the block_reader memory user (mu)
	bw_idle           : in std_logic; --interface to the block_writer memory user (mu)
	bs_idle           : in std_logic; --interface to the block_set memory user
	ps_idle           : in  std_logic --interface to the polyscanline mu
	);	 
end eppcu;

architecture Behavioral of eppcu is
	--the epp host (computer) writes an address, then next writes a data byte to go into the desired address's register
	signal epp_address : std_logic_vector(7 downto 0);

	signal latched_epp_write 				: std_logic;
	signal latched_epp_data_strobe 		: std_logic;
	signal latched_epp_address_strobe 	: std_logic;
	signal epp_error : std_logic;

	--registers that hold data for read/write from the epp port
	signal epp_dr_0 : std_logic_vector(7 downto 0);
	signal epp_dr_1 : std_logic_vector(7 downto 0);
	signal epp_dr_2 : std_logic_vector(7 downto 0);
	signal epp_dr_3 : std_logic_vector(7 downto 0);
	signal epp_dr_4 : std_logic_vector(7 downto 0);

	--the epp port is slow compared to the rest of the fpga, so we do a lot of waiting
	signal wait_timer : std_logic_vector(11 downto 0);
	type EPP_INT_STATE is ( EPP_NORMAL, EPP_DELAY_HOST_ADDRESS_WRITE_START, EPP_DELAY_HOST_DATA_WRITE_START, EPP_DELAY_NOTIFY_HOST_DATA_READ,
                           EPP_ENDING_HOST_DATA_READ, EPP_ENDING_HOST_DATA_WRITE, EPP_ENDING_HOST_ADDRESS_WRITE,
									EPP_CB_WRITE, EPP_CB_READ, EPP_SIMPLE_DELAY);
	signal epp_state : EPP_INT_STATE;

	--when in state EPP_SIMPLE_DELAY we wait until wait_timer == simple_delay_time
	--then set our epp_state to be simple_delay_post_state
	signal simple_delay_post_state 			: EPP_INT_STATE;
 	signal simple_delay_time 					: std_logic_vector(11 downto 0);

	procedure BeginSimpleDelay(
		constant state_after_delay 	: in EPP_INT_STATE;
		constant des_delay_time 		: in std_logic_vector(11 downto 0);
		signal epp_state 						: out EPP_INT_STATE;
		signal simple_delay_post_state 	: out EPP_INT_STATE;
		signal simple_delay_time 			: out std_logic_vector(11 downto 0);
		variable reset_wait_timer			: out boolean ) is
	begin
		simple_delay_post_state <= state_after_delay;
		simple_delay_time <= des_delay_time;
		epp_state <= EPP_SIMPLE_DELAY;
		reset_wait_timer := true;
	end BeginSimpleDelay;

begin

   cmd_bus_data(31 downto 24) <= epp_dr_0;
   cmd_bus_data(23 downto 16) <= epp_dr_1;
   cmd_bus_data(15 downto 8)  <= epp_dr_2;
   cmd_bus_data(7 downto 0)   <= epp_dr_3;

   cmd_bus_addr <= epp_dr_4;

	process(tvc_clock)
	variable ending_host_data_write : boolean;
	variable reset_wait_timer : boolean;
   variable inc_epp_address : boolean;
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		epp_error <= '0';
		epp_address <= X"00";

		epp_data <= (others => 'Z');
		epp_wait <= '0';

      cmd_bus_strobe <= '0';
		
		epp_state <= EPP_NORMAL;
		--simple_delay_post_state <= EPP_NORMAL;
		
		wait_timer <= X"000";
		--simple_delay_time <= X"000";
		
		epp_irq <= '0';		
	else	--not a reset
		ending_host_data_write := false;
		reset_wait_timer := false;
      inc_epp_address := false;
		
		latched_epp_write <= epp_write;
      latched_epp_data_strobe <= epp_data_strobe; 
      latched_epp_address_strobe <= epp_address_strobe;
		
	case epp_state is
	 when EPP_ENDING_HOST_DATA_READ =>
		--before we enter this state we have already waited X"014" ticks, so we are immediatly ready
		--to check for latched_epp_data_strobe to be high
		if latched_epp_data_strobe = '1' then
			epp_data <= (others => 'Z');
			epp_state <= EPP_NORMAL;
			epp_wait <= '0';
		elsif wait_timer = X"1f4" then  --a ?hackish? timeout
			epp_data <= (others=> 'Z');
			epp_state <= EPP_NORMAL;
			epp_wait <= '0';
		end if;
	when EPP_ENDING_HOST_DATA_WRITE =>
		--before we enter this state we have already waited X"014" ticks, so we are immediatly ready
		--to check for latched_epp_data_strobe to be high
		if latched_epp_data_strobe = '1' then
			epp_wait <= '0';
			epp_state <= EPP_NORMAL;
			epp_data <= (others => 'Z');
		elsif wait_timer = X"1f4" then --timeout
			epp_wait <= '0';
			epp_state <= EPP_NORMAL;
			epp_data <= (others => 'Z');
		end if;
	when EPP_ENDING_HOST_ADDRESS_WRITE =>
		--before we enter this state we have already waited X"014" ticks, so we are immediatly ready
		--to check for latched_epp_data_strobe to be high
		if latched_epp_address_strobe = '1' then
			epp_wait <= '0';
			epp_state <= EPP_NORMAL;
			epp_data <= (others => 'Z');
		elsif wait_timer = X"1f4" then --timeout
			epp_wait <= '0';
			epp_state <= EPP_NORMAL;
			epp_data <= (others => 'Z');
		end if;
	when EPP_NORMAL =>
		if latched_epp_address_strobe = '0' and latched_epp_write = '0' then
			BeginSimpleDelay(EPP_DELAY_HOST_ADDRESS_WRITE_START, X"010", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
		elsif latched_epp_data_strobe = '0' and latched_epp_write = '0' then
			BeginSimpleDelay(EPP_DELAY_HOST_DATA_WRITE_START, X"010", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
		elsif latched_epp_data_strobe = '0' and latched_epp_write = '1' then
		   case epp_address is
				when X"00" =>
					epp_data <= epp_dr_0;
				when X"01" =>
					epp_data <= epp_dr_1;
				when X"02" =>
					epp_data <= epp_dr_2;
				when X"03" =>
					epp_data <= epp_dr_3;
				when X"04" =>
					epp_data <= epp_dr_4;
				when X"05" =>
					epp_data <= X"00";
				when X"06" =>
					epp_data <= epp_error & br_idle & bw_idle & bs_idle & ps_idle & "000";
				when others =>
					epp_data <= X"00";
			end case;
			BeginSimpleDelay(EPP_DELAY_NOTIFY_HOST_DATA_READ, X"010", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
		end if; --done read and write
	when EPP_DELAY_HOST_ADDRESS_WRITE_START =>
	  	epp_address <= epp_data;
		epp_wait <= '1';
		BeginSimpleDelay(EPP_ENDING_HOST_ADDRESS_WRITE, X"010", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
	when	EPP_DELAY_HOST_DATA_WRITE_START =>
		ending_host_data_write := true;
      inc_epp_address := true;
		case epp_address is
		when X"00" =>
			epp_dr_0 <= epp_data;
		when X"01" =>
			epp_dr_1 <= epp_data;
		when X"02" =>
			epp_dr_2 <= epp_data;
		when X"03" =>
			epp_dr_3 <= epp_data;
		when X"04" => --the command bus address
			epp_dr_4 <= epp_data;
         inc_epp_address := false;
		when X"05" =>
         inc_epp_address := false;
         if epp_data = X"F0" then --do a command bus write
				ending_host_data_write := false;
            epp_state <= EPP_CB_WRITE;
            cmd_bus_strobe <= '1';
			elsif epp_data = X"0F" then --do a command read
            ending_host_data_write := false;
            epp_state <= EPP_CB_READ;
            --cmd_bus_strobe <= '1'; --only one value exists!
         end if; --otherwise we do a ending_host_data_write
		when others => --should never happen (invalid address in address register)
         --ending_host_data_write := true;
         null;
		end case;
		if ending_host_data_write = true then
			epp_wait <= '1';
			BeginSimpleDelay(EPP_ENDING_HOST_DATA_WRITE, X"014", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
		end if;
	when EPP_DELAY_NOTIFY_HOST_DATA_READ=>
		epp_wait <= '1';
		BeginSimpleDelay(EPP_ENDING_HOST_DATA_READ, X"014", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
   when EPP_CB_WRITE =>
      cmd_bus_strobe <= '0';
      epp_wait <= '1';
		BeginSimpleDelay(EPP_ENDING_HOST_DATA_WRITE, X"014", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
   when EPP_CB_READ =>
      epp_dr_0 <= cmd_bus_data_out(31 downto 24);
      epp_dr_1 <= cmd_bus_data_out(23 downto 16);
      epp_dr_2 <= cmd_bus_data_out(15 downto 8);
      epp_dr_3 <= cmd_bus_data_out(7 downto 0);
      --cmd_bus_strobe <= '0';
      epp_wait <= '1';
		BeginSimpleDelay(EPP_ENDING_HOST_DATA_WRITE, X"014", epp_state, simple_delay_post_state, simple_delay_time, reset_wait_timer);
	when EPP_SIMPLE_DELAY =>
		if wait_timer = simple_delay_time then
			reset_wait_timer := true;
			epp_state <= simple_delay_post_state;
		end if;
	end case;
	
	if reset_wait_timer = true then
		wait_timer <= X"000";
	else
		wait_timer <= wait_timer + 1;
	end if;
	
   if inc_epp_address = true then
      epp_address <= epp_address + 1;
   end if;
   
  	end if; --reset
	end if; --clock
	end process;
end Behavioral;
