library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity pix_pusher is
    Port (
		mu_bus_address 	: in std_logic_vector(11 downto 0);
		mu_bus_data		: inout std_logic_vector(31 downto 0);
		mu_bus_write		: in std_logic;
		mu_bus_strobe		: in std_logic;

		odd_sc_irq			: out std_logic;
		odd_sc_select		: in std_logic;
		even_sc_irq			: out std_logic;
		even_sc_select		: in std_logic;

    	tvc_clock 			: in 	std_logic;
		tvc_reset 			: in 	std_logic;

		disp_pix_value 	: out 	std_logic_vector(7 downto 0);
		disp_horiz_sync 	: out 	std_logic;
		disp_vert_sync 	: out 	std_logic
		);
end pix_pusher;

architecture Behavioral of pix_pusher is
	type PP_INT_STATE is (PP_PRE_HALT, PP_NORMAL, PP_ODD_FILL, PP_EVEN_FILL);
	signal pp_state : PP_INT_STATE;

	--signal framebuffer_start_addr	: std_logic_vector(23 downto 0);

	signal dd_reset_cntrs_req			: std_logic;

	signal dd_even_fill_req				: std_logic;
	signal even_sc_start_fill			: std_logic;
	signal even_sc_start_fill_addr	: std_logic_vector(23 downto 0);

	signal dd_odd_fill_req				: std_logic;
	signal odd_sc_start_fill			: std_logic;
	signal odd_sc_start_fill_addr		: std_logic_vector(23 downto 0);

	--these signals are operating in the pixel clock domain and connect to the display driver.
	signal pixel_clock	:  std_logic;
	signal pix_addr_sig	:  std_logic_vector(11 downto 0);
	signal odd_pix_sig	:  std_logic_vector(31 downto 0);
	signal even_pix_sig	:  std_logic_vector(31 downto 0);

	component scanline_cache is
	Port(
	mu_bus_address 		: in 		std_logic_vector(11 downto 0);
	mu_bus_data				: inout 	std_logic_vector(31 downto 0);
	mu_bus_write			: in		std_logic;
	mu_bus_strobe			: in  	std_logic;
	sc_irq					: out		std_logic;
	sc_select				: in		std_logic;
	tvc_clock 				: in		std_logic;
	tvc_reset 				: in		std_logic;
	sc_start_fill			: in		std_logic;
	sc_start_fill_addr	: in		std_logic_vector(23 downto 0); 
	pixel_clock				: in     std_logic;
	next_pix_addr			: in     std_logic_vector(11 downto 0);
	next_pix_values		: out    std_logic_vector(31 downto 0)
	);
	end component;

	component disp_driver is
	Port (
	tvc_clock			: in	std_logic;
	tvc_reset			: in	std_logic; 
	odd_fill_req		: out std_logic;
	even_fill_req		: out std_logic;
	reset_cntrs_req	: out std_logic;
	pixel_clock 		: in  std_logic;
	next_pix_addr		: out	std_logic_vector(11 downto 0);
	odd_next_pix_val	: in	std_logic_vector(31 downto 0);
	even_next_pix_val	: in 	std_logic_vector(31 downto 0);
	disp_pix_value 	: out std_logic_vector(7 downto 0);
	disp_horiz_sync 	: out std_logic;
	disp_vert_sync 	: out std_logic
	);
	end component;
	
	begin

	odd_sc_sc : scanline_cache
		port map(mu_bus_address=>mu_bus_address, mu_bus_data=>mu_bus_data, mu_bus_write=>mu_bus_write, mu_bus_strobe=>mu_bus_strobe,
					sc_irq=>odd_sc_irq, sc_select=>odd_sc_select, tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
					sc_start_fill=>odd_sc_start_fill, sc_start_fill_addr=>odd_sc_start_fill_addr, 
					pixel_clock=>pixel_clock, next_pix_addr=>pix_addr_sig, next_pix_values=>odd_pix_sig );

	even_sc_sc : scanline_cache
		port map(mu_bus_address=>mu_bus_address, mu_bus_data=>mu_bus_data, mu_bus_write=>mu_bus_write, mu_bus_strobe=>mu_bus_strobe,
					sc_irq=>even_sc_irq, sc_select=>even_sc_select, tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,	
					sc_start_fill=>even_sc_start_fill, sc_start_fill_addr=>even_sc_start_fill_addr, 
					pixel_clock=>pixel_clock, next_pix_addr=>pix_addr_sig, next_pix_values=>even_pix_sig );

	d_driver : disp_driver
		port map(tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
					odd_fill_req=>dd_odd_fill_req, even_fill_req=>dd_even_fill_req, reset_cntrs_req=>dd_reset_cntrs_req,
					pixel_clock=>pixel_clock, next_pix_addr=>pix_addr_sig,
					odd_next_pix_val=>odd_pix_sig, even_next_pix_val=>even_pix_sig,
					disp_pix_value=>disp_pix_value, disp_horiz_sync=>disp_horiz_sync, disp_vert_sync=>disp_vert_sync);

	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		pp_state <= PP_PRE_HALT;
		--framebuffer_start_addr <= X"000000";
		even_sc_start_fill	<= '0';
		even_sc_start_fill_addr	<= X"000000";
		odd_sc_start_fill <= '0';
		odd_sc_start_fill_addr <= X"000000";
	else	--not a reset
	
		if pixel_clock = '1' then
		  pixel_clock <= '0';
		 else
			pixel_clock <= '1';
		end if;

		case pp_state is
		when PP_PRE_HALT=>	--null; --debugging trap here
			pp_state <= PP_NORMAL;
		when PP_NORMAL =>
			if dd_reset_cntrs_req = '1' then
				odd_sc_start_fill_addr <=  X"000280";
				even_sc_start_fill_addr <= X"000000";
			elsif dd_odd_fill_req = '1' then
				odd_sc_start_fill <= '1';
				pp_state <= PP_ODD_FILL;
			elsif	dd_even_fill_req = '1' then
				even_sc_start_fill <= '1';
				pp_state <= PP_EVEN_FILL;
			end if;
		when PP_ODD_FILL =>
			odd_sc_start_fill <= '0';
			odd_sc_start_fill_addr <= odd_sc_start_fill_addr + 1280;
			pp_state <= PP_NORMAL;
		when PP_EVEN_FILL =>
			even_sc_start_fill <= '0';
			even_sc_start_fill_addr <= even_sc_start_fill_addr + 1280;
			pp_state <= PP_NORMAL;
		end case;

	end if; --reset
	end if; --clock
	end process;

end Behavioral;
